#pragma once
#include "TProcuraMelhorativa.h"

///////////////////////////////////////////////////////////////////////////////
//	CDamas class
///////////////////////////////////////////////////////////////////////////////
//	Author: Jose Coelho
//	Description:
//    e-flio B de 21/22
//    Este e-flio constri no e-flio A, existindo de igual forma um tabuleiro de NxN, 
//    mas pretende-se colocar torres no tabuleiro. As torres movem-se apenas na horizontal e vertical, 
//    mas ao contrrio do xadrez, tm 3 cores. Pretende-se a colocao do maior nmero de torres, 
//    sem que duas torres da mesma cor se ataquem mutuamente. 
// 
//    - prespetiva construtiva: 
//      - manter um tabuleiro com as posies das torres, e possiblidades de uso das casas vazias
//      - inverter o valor do estado, para melhor permitir a utilizao do A*: 3N-#torres
//      - cada movimento  uma seleo de uma pea possvel para uma casa (um movimento pode anular outro)
//        - inicialmente ser NxNx3 mas com as simetrias quase todas os os vizinhos so anulados
//        - manter a ltima pea colocada, de modo a fixar uma ordem de colocao de peas
//      - simetrias: 
//        - os 3 eixos considerando torres sem cor, e de seguida ordenar as cores por uma ordem 
//      - heurstica: nmero de casas com possibilidade de ter uma torre
//      - heurstica 2: segundo critrio,  o nmero de torres distintas possveis no total das casas
// 
//    - prespetiva melhorativa: 
//      - gerador aleatrio, com uma permutao de todas as casas/possibilidades, colocando as que so possveis
//      - vizinhana/mutao: colocar mais torres nas posies vazias, e retirar torres em violao
//      - cruzamento: colocar torres alternativamente do pai/me, se forem vlidas. Se no final ainda existirem
//        casas vlidas, preencher aleatoriamente
// 
///////////////////////////////////////////////////////////////////////////////

// 2 primeiros bits
#define _TORRE(casa) (casa&3)
// 3 bits seguintes
#define _ALTTORRE(casa) ((casa>>2)&7)
// set 2 primeiros bits
#define _SETTORRE(casa,value) casa=(casa&~3)|(value&3)
// set 3 bits seguintes
#define _SETALTTORRE(casa,value) casa=casa&(~(7<<2))|((value&7)<<2)

//#define HASHSIZE 999983


class CTorres :
	public TProcuraMelhorativa
{
private:
	// acesso ao tabuleiro
	inline int &Casa(int linha, int coluna) { return tabuleiro[linha * N + coluna]; }
	bool ColocarTorre(int linha, int coluna, int cor); // cores: 1, 2, 3 
	void Normalizar(); // normalizar este tabuleiro, antes de utilizar a hashtable
	int Peso(); // peso de um tabuleiro, para colocar na verso normalizada de menor peso
	void Inverter(int eixo); // inverte o tabuleiro no eixo especificado
	unsigned int hash(); 
public:
	CTorres(void);
	~CTorres(void);

	void Completar(); // preencher o resto das casas pendentes de forma aleatria

	// dados de instncia
	static int N;
#ifdef HASHSIZE
	static TVector<int> hashTable[HASHSIZE];
	static void ResetHashTable();
	static int hashTableID(unsigned int hash, TVector<int>& tabuleiro); // adiciona se no existe, ou identifica que j existe (retorna -1)
#endif

	// dados de estado
	TVector<int> tabuleiro; // torres em cada posio do tabuleiro, e trs bits com as possibilidades de colocao
	int torres, inutilizadas; // nmero de torres colocadas e nmero de casas inutilizadas
	int ultLinha, ultColuna; // ltima pea a ser colocada, para gerar sucessores sem ser em ciclo

	// metodos redefinidos de TProcuraConstrutiva

	// Cria um objecto que e uma copia deste
	TProcuraConstrutiva* Duplicar(void);
	void Copiar(TProcuraConstrutiva*objecto) {
		TProcuraMelhorativa::Copiar(objecto);
		torres=((CTorres*)objecto)->torres;
		inutilizadas = ((CTorres*)objecto)->inutilizadas;
		tabuleiro = ((CTorres*)objecto)->tabuleiro;
		ultLinha = ((CTorres*)objecto)->ultLinha;
		ultColuna = ((CTorres*)objecto)->ultColuna;
	}
	// Coloca o objecto no estado inicial da procura
	void SolucaoVazia(void);
	// Coloca em sucessores a lista de objectos sucessores (sao alocados neste metodo e tem de ser apagados)
	// O custo nao necessita de ser preenchido, caso seja sempre unitario
	void Sucessores(TVector<TProcuraConstrutiva*>&sucessores, TVector<int>&custo);
	// Retorna verdade caso o estado actual seja um estado objectivo
	bool SolucaoCompleta(void);
	// Escrever informacao de debug sobre o objecto currente
	// (utilizar variavel TProcuraConstrutiva::debug para seleccionar o detalhe pretendido)
	void Debug(void);

	int Heuristica(void);

	// metodos redefinidos de TProcuraMelhorativa

	// e gerada uma nova solucao aleatoriamente.
	void NovaSolucao(void);
	// Retorna o valor da solucao completa actual.
	int Avaliar(void);
	// Operador de vizinhanca (apenas em solucaes completas)
	// chamar este operador para actualizacao de estatisticas
	void Vizinhanca(TVector<TProcuraMelhorativa*>&vizinhos);
	// muta o estado actual (de acordo com a parametrizacao global)
	void Mutar(void);
	// cruza os dois elementos a e b, colocando o resultado neste
	void Cruzamento(TProcuraMelhorativa*a,TProcuraMelhorativa*b);
	// retorna a distancia deste elemento ao elemento a
	// (necessario para manter os elementos diversos)
	int Distancia(TProcuraMelhorativa*a);


	void LimparEstatisticas(void) {
#ifdef HASHSIZE
		ResetHashTable();
#endif
		TProcuraConstrutiva::LimparEstatisticas();
	}

};
